# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""
Extension points for Quod Libet and Ex Falso

This (currently) contains code plugins (formerly kept in /plugins)
arranged by plugin type (songsmenu, events, etc).

Base classes themselves are in /quodlibet/plugins
"""
